from test.testbase import DbIsolatedTest, run_as_sync

from db.python.layers.family import FamilyLayer
from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.search import SearchLayer
from db.python.layers.sequencing_group import SequencingGroupLayer
from db.python.tables.family_participant import FamilyParticipantTable
from models.enums import SearchResponseType
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AssayUpsertInternal,
    FamilySearchResponseData,
    ParticipantSearchResponseData,
    ParticipantUpsertInternal,
    PedRowInternal,
    SampleSearchResponseData,
    SampleUpsertInternal,
    SequencingGroupSearchResponseData,
    SequencingGroupUpsertInternal,
)
from models.models.sample import sample_id_format
from models.models.sequencing_group import sequencing_group_id_format


class TestSample(DbIsolatedTest):
    """Test sample class"""

    # tests run in 'sorted by ascii' order
    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        self.schlay = SearchLayer(self.connection)
        self.slayer = SampleLayer(self.connection)
        self.player = ParticipantLayer(self.connection)
        self.flayer = FamilyLayer(self.connection)
        self.sglayer = SequencingGroupLayer(self.connection)

    @run_as_sync
    async def test_search_non_existent_sample_by_internal_id(self):
        """
        Search by CPG sample ID that doesn't exist
        """
        cpg_id = sample_id_format(9_999_999_999)
        results = await self.schlay.search(query=cpg_id, project_ids=[self.project_id])
        self.assertEqual(0, len(results))

    @run_as_sync
    async def test_search_unavailable_sample_by_internal_id(self):
        """
        Search by CPG sample ID that you do not have access to
        Mock this in testing by limiting scope to non-existent project IDs
        """
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX001'}, type='blood'
            )
        )
        cpg_id = sample_id_format(sample.id)

        results = await self.schlay.search(
            query=cpg_id, project_ids=[self.project_id + 1]
        )
        self.assertEqual(1, len(results))

    @run_as_sync
    async def test_search_isolated_sample_by_id(self):
        """
        Search by valid CPG sample ID (special case)
        """
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX001'}, type='blood'
            )
        )
        cpg_id = sample_id_format(sample.id)
        results = await self.schlay.search(query=cpg_id, project_ids=[self.project_id])

        self.assertEqual(1, len(results))
        self.assertEqual(cpg_id, results[0].title)
        self.assertEqual(cpg_id, results[0].data.id)

        result_data = results[0].data
        self.assertIsInstance(result_data, SampleSearchResponseData)
        assert isinstance(result_data, SampleSearchResponseData)
        self.assertListEqual(['EX001'], result_data.sample_external_ids)

    @run_as_sync
    async def test_search_isolated_sequencing_group_by_id(self):
        """
        Search by valid CPG sequencing group ID (special case)
        """
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EXS001'}, type='blood'
            )
        )
        sg = await self.sglayer.upsert_sequencing_groups(
            [
                SequencingGroupUpsertInternal(
                    sample_id=sample.id,
                    technology='long-read',
                    platform='illumina',
                    meta={
                        'sequencing_type': 'transcriptome',
                        'sequencing_technology': 'long-read',
                        'sequencing_platform': 'illumina',
                    },
                    type='transcriptome',
                    assays=[
                        AssayUpsertInternal(
                            type='sequencing',
                            meta={
                                'sequencing_type': 'transcriptome',
                                'sequencing_technology': 'long-read',
                                'sequencing_platform': 'illumina',
                            },
                        )
                    ],
                )
            ]
        )
        cpg_sg_id = sequencing_group_id_format(sg[0].id)
        results = await self.schlay.search(
            query=cpg_sg_id, project_ids=[self.project_id]
        )

        self.assertEqual(1, len(results))
        self.assertEqual(cpg_sg_id, results[0].title)
        result_data = results[0].data
        assert isinstance(result_data, SequencingGroupSearchResponseData)
        self.assertIsInstance(result_data, SequencingGroupSearchResponseData)
        self.assertEqual(cpg_sg_id, result_data.id)
        self.assertEqual(cpg_sg_id, result_data.sg_external_id)

    @run_as_sync
    async def test_search_isolated_sample_by_external_id(self):
        """
        Search by External sample ID with no participant / family,
        should only return one result
        """
        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'EX001'}, type='blood'
            )
        )
        results = await self.schlay.search(query='EX001', project_ids=[self.project_id])

        cpg_id = sample_id_format(sample.id)

        self.assertEqual(1, len(results))

        self.assertEqual(cpg_id, results[0].title)
        result_data = results[0].data

        self.assertIsInstance(result_data, SampleSearchResponseData)
        assert isinstance(result_data, SampleSearchResponseData)
        self.assertEqual(cpg_id, result_data.id)
        self.assertListEqual(['EX001'], result_data.sample_external_ids)
        self.assertListEqual([], result_data.participant_external_ids)
        self.assertListEqual([], result_data.family_external_ids)

    @run_as_sync
    async def test_search_participant_isolated(self):
        """
        Search participant w/ no family by External ID
        should only return one result
        """
        p = await self.player.upsert_participant(
            ParticipantUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'PART01'})
        )
        results = await self.schlay.search(
            query='PART01', project_ids=[self.project_id]
        )
        self.assertEqual(1, len(results))

        self.assertEqual('PART01', results[0].title)
        result_data = results[0].data
        assert isinstance(result_data, ParticipantSearchResponseData)
        self.assertEqual(p.id, result_data.id)
        self.assertListEqual(['PART01'], result_data.participant_external_ids)
        self.assertListEqual([], result_data.family_external_ids)

    @run_as_sync
    async def test_search_family(self):
        """
        Search family by External ID
        should only return one result
        """
        f_id = await self.flayer.create_family(external_id='FAMXX01')
        results = await self.schlay.search(
            query='FAMXX01', project_ids=[self.project_id]
        )
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual('FAMXX01', result.title)
        result_data = result.data
        assert isinstance(result_data, FamilySearchResponseData)
        self.assertEqual(f_id, result_data.id)
        self.assertListEqual(['FAMXX01'], result_data.family_external_ids)

    @run_as_sync
    async def test_search_mixed(self):
        """Create a number of resources, and search for all of them"""
        fptable = FamilyParticipantTable(self.connection)

        p = await self.player.upsert_participant(
            ParticipantUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'X:PART01'})
        )
        f_id = await self.flayer.create_family(external_id='X:FAM01')
        await fptable.create_rows(
            [
                PedRowInternal(
                    family_id=f_id,
                    individual_id=p.id,
                    paternal_id=None,
                    maternal_id=None,
                    affected=0,
                    sex=0,
                    notes=None,
                )
            ]
        )

        sample = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'X:SAM001'},
                type='blood',
                participant_id=p.id,
            )
        )

        all_results = await self.schlay.search(
            query='X:', project_ids=[self.connection.project_id]
        )
        self.assertEqual(3, len(all_results))

        family_result = next(
            r for r in all_results if r.type == SearchResponseType.FAMILY
        )
        participant_result = next(
            r for r in all_results if r.type == SearchResponseType.PARTICIPANT
        )
        sample_result = next(
            r for r in all_results if r.type == SearchResponseType.SAMPLE
        )
        family_result_data = family_result.data
        participant_result_data = participant_result.data
        sample_result_data = sample_result.data

        assert isinstance(family_result_data, FamilySearchResponseData)
        assert isinstance(participant_result_data, ParticipantSearchResponseData)
        assert isinstance(sample_result_data, SampleSearchResponseData)

        # linked family matches
        self.assertEqual('X:FAM01', family_result.title)

        # linked participant matches
        self.assertEqual('X:PART01', participant_result.title)
        self.assertListEqual(['X:FAM01'], participant_result_data.family_external_ids)

        # linked sample matches
        cpg_id = sample_id_format(sample.id)
        self.assertEqual(cpg_id, sample_result_data.id)
        self.assertListEqual(['X:SAM001'], sample_result_data.sample_external_ids)
        self.assertListEqual(['X:FAM01'], participant_result_data.family_external_ids)
        self.assertListEqual(
            ['X:PART01'], participant_result_data.participant_external_ids
        )



# pylint: disable=no-member
import csv
import logging
import os
from collections import defaultdict
from typing import Any

from cloudpathlib import AnyPath

from cpg_utils.cloud import get_path_components_from_gcp_path

from metamist.parser.cloudhelper import CloudHelper


class AuditHelper(CloudHelper):
    """General helper class for bucket auditing"""

    EXCLUDED_SGS: set[str] = set(
        sg for sg in os.getenv('SM_AUDIT_EXCLUDED_SGS', '').split(',') if sg
    )

    @staticmethod
    def get_gcs_bucket_subdirs_to_search(paths: list[str]) -> defaultdict[str, list]:
        """
        Takes a list of paths and extracts the bucket name and subdirectory, returning all unique pairs
        of buckets/subdirectories
        """
        buckets_subdirs_to_search: defaultdict[str, list] = defaultdict(list)
        for path in paths:
            try:
                pc = get_path_components_from_gcp_path(path)
            except ValueError:
                logging.warning(f'{path} invalid')
                continue
            bucket = pc['bucket']
            subdir = pc['suffix']
            if subdir and subdir not in buckets_subdirs_to_search[bucket]:
                buckets_subdirs_to_search[bucket].append(subdir)

        return buckets_subdirs_to_search

    def get_gcs_paths_for_subdir(
        self, bucket_name: str, subdirectory: str, file_extension: tuple[str]
    ):
        """Iterate through a gcp bucket/subdir and get all the blobs with the specified file extension(s)"""
        files_in_bucket_subdir = []
        for blob in self.gcs_client.list_blobs(
            bucket_name, prefix=subdirectory, delimiter='/'
        ):
            # Check if file ends with specified analysis type
            if not blob.name.endswith(file_extension):
                continue
            files_in_bucket_subdir.append(f'gs://{bucket_name}/{blob.name}')

        return files_in_bucket_subdir

    def find_files_in_gcs_buckets_subdirs(
        self, buckets_subdirs: defaultdict[str, list], file_types: tuple[str]
    ):
        """
        Takes a list of (bucket,subdirectory) tuples and finds all the files contained in that directory
        with filetypes defined with an input list
        """
        files_in_bucket = []
        for bucket, subdirs in buckets_subdirs.items():
            for subdir in subdirs:
                # matrixtable / hailtable subdirectories should not appear in main-upload buckets,
                # but handle them just in case. These directories are too large to search.
                if '.mt' in subdir or '.ht' in subdir:
                    continue
                files_in_bucket.extend(
                    self.get_gcs_paths_for_subdir(bucket, subdir, file_types)
                )

        return files_in_bucket

    def find_assay_files_in_gcs_bucket(
        self, bucket_name: str, file_extensions: tuple[str]
    ) -> list[str]:
        """Gets all the gs paths to fastq files in the datasets upload bucket"""
        if bucket_name.startswith('gs://'):
            bucket_name = bucket_name.removeprefix('gs://')
        assay_paths = []
        if 'upload' not in bucket_name:
            # No prefix means it will get all blobs in the bucket (regardless of path)
            # This can be a dangerous call outside of the upload buckets
            raise NameError(
                'Call to list_blobs without prefix only valid for upload buckets'
            )

        for blob in self.gcs_client.list_blobs(bucket_name, prefix=''):
            if blob.name.endswith(file_extensions):
                assay_paths.append(f'gs://{bucket_name}/{blob.name}')
            continue

        return assay_paths

    @staticmethod
    def get_sequencing_group_ids_from_analysis(analysis) -> list[str]:
        """Tries a number of different field names to retrieve the sg ids from an analysis"""
        while True:
            try:
                sg_ids = analysis['meta']['sample']
                break
            except KeyError:
                pass

            try:
                sg_ids = analysis['meta']['samples']
                break
            except KeyError:
                pass

            try:
                sg_ids = analysis['meta']['sample_ids']
                break
            except KeyError:
                pass

            try:
                sg_ids = analysis['meta']['sequencing_group']
                break
            except KeyError:
                pass

            try:
                sg_ids = analysis['meta']['sequencing_groups']
                break
            except KeyError as exc:
                raise ValueError(
                    f'Analysis {analysis["id"]} missing sample or sequencing group field.'
                ) from exc

        if isinstance(sg_ids, str):
            return [
                sg_ids,
            ]
        return sg_ids

    @staticmethod
    def write_csv_report_to_cloud(
        data_to_write: list[Any], report_path: AnyPath, header_row: list[str] | None
    ):
        """
        Writes a csv report to the cloud bucket containing the data to write
        at the report path, with an optional header row
        """
        with AnyPath(report_path).open('w+') as f:  # pylint: disable=E1101
            writer = csv.writer(f)
            if header_row:
                writer.writerow(header_row)
            for row in data_to_write:
                if isinstance(row, str):
                    writer.writerow([row])
                    continue
                writer.writerow(row)

        logging.info(f'Wrote {len(data_to_write)} lines to report: {report_path}')



import logging
import os
from collections import defaultdict, namedtuple
from datetime import datetime
from functools import cache
from typing import Any

from gql.transport.requests import log as requests_logger

from metamist.audit.audithelper import AuditHelper
from metamist.graphql import gql, query_async

ANALYSIS_TYPES_QUERY = gql(
    """
    query analysisTypes {
        enum {
            analysisType
        }
    }
    """
)

QUERY_PARTICIPANTS_SAMPLES_SGS_ASSAYS = gql(
    """
        query DatasetData($datasetName: String!) {
            project(name: $datasetName) {
                participants {
                    id
                    externalId
                    samples {
                        id
                        externalId
                        sequencingGroups {
                            id
                            type
                            assays {
                                id
                                meta
                            }
                        }
                    }
                }
            }
        }
    """
)

QUERY_SG_ANALYSES = gql(
    """
        query sgAnalyses($dataset: String!, $sgIds: [String!], $analysisTypes: [String!]) {
          sequencingGroups(id: {in_: $sgIds}, project: {eq: $dataset}) {
            id
            analyses(status: {eq: COMPLETED}, type: {in_: $analysisTypes}, project: {eq: $dataset}) {
              id
              meta
              output
              type
              timestampCompleted
            }
          }
        }
    """
)

# Variable type definitions
AssayId = int
SampleId = str
SampleExternalId = str
ParticipantExternalId = str

AssayReportEntry = namedtuple(
    'AssayReportEntry',
    'sg_id assay_id assay_file_path analysis_id filesize',
)


@cache
async def get_analysis_types():
    """Return the list of analysis types from the enum table."""
    logging.getLogger().setLevel(logging.WARN)
    analysis_types = await query_async(ANALYSIS_TYPES_QUERY)
    logging.getLogger().setLevel(logging.INFO)
    return analysis_types['enum'][  # pylint: disable=unsubscriptable-object
        'analysisType'
    ]


class GenericAuditor(AuditHelper):
    """Auditor for cloud storage buckets"""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        dataset: str,
        sequencing_types: list[str],
        file_types: tuple[str],
        default_analysis_type='cram',
        default_analysis_status='completed',
    ):
        if not dataset:
            raise ValueError('Metamist dataset is required')

        super().__init__(search_paths=None)

        self.dataset = dataset
        self.sequencing_types = sequencing_types
        self.file_types = file_types
        self.default_analysis_type: str = default_analysis_type
        self.default_analysis_status: str = default_analysis_status

        requests_logger.setLevel(logging.WARNING)

    async def get_participant_data_for_dataset(self) -> list[dict]:
        """
        Uses a graphQL query to return all participants in a Metamist dataset.
        Returned list includes all samples and assays associated with the participants.
        """

        logging.getLogger().setLevel(logging.WARN)
        participant_query_results = await query_async(
            QUERY_PARTICIPANTS_SAMPLES_SGS_ASSAYS, {'datasetName': self.dataset}
        )
        logging.getLogger().setLevel(logging.INFO)

        participant_data = participant_query_results['project']['participants']

        filtered_participants = []
        for participant in participant_data:
            if not participant['samples']:
                logging.info(
                    f'{self.dataset} :: Filtering participant {participant["id"]} ({participant["externalId"]}) as it has no samples.'
                )
                continue
            filtered_participants.append(participant)

        return filtered_participants

    @staticmethod
    def get_most_recent_analyses_by_sg(
        analyses_list: list[dict[str, Any]],
    ) -> dict[str, dict[str, Any]]:
        """
        Takes a list of completed analyses for a number of sequencing groups and returns the latest
        completed analysis for each sequencing group, creating a 1:1 mapping of SG to analysis.
        """
        most_recent_analysis_by_sg = {}

        for sg_analyses in analyses_list:
            sg_id = sg_analyses['id']
            analyses = sg_analyses['analyses']
            if not analyses:
                continue
            if len(analyses) == 1:
                most_recent_analysis_by_sg[sg_id] = analyses[0]
                continue

            sorted_analyses = sorted(
                analyses,
                key=lambda x: datetime.strptime(
                    x['timestampCompleted'], '%Y-%m-%dT%H:%M:%S'
                ),
            )
            most_recent_analysis_by_sg[sg_id] = sorted_analyses[-1]

        return most_recent_analysis_by_sg

    @staticmethod
    def map_internal_to_external_sample_ids(
        participants: list[dict],
    ) -> dict[SampleId, SampleExternalId]:
        """
        Returns the {internal sample id : external sample id} mapping for all participants in a Metamist dataset
        """
        return {
            sample['id']: sample['externalId']
            for participant in participants
            for sample in participant['samples']
        }

    def get_assay_map_from_participants(
        self, participants: list[dict]
    ) -> tuple[dict[str, str], dict[int, str], dict[Any, list[tuple[Any, Any]]]]:
        """
        Input the list of Metamist participant dictionaries from the 'QUERY_PARTICIPANTS_SAMPLES_SGS_ASSAYS' query

        Returns the mappings:
        1. { sg_id       : sample_id }
        2. { assay_id    : sg_id }
        3. { assay_id    : (read_filepath, read_filesize,) }
        """

        sg_sample_id_map = {}
        assay_sg_id_map = {}
        assay_filepaths_filesizes = defaultdict(list)

        sample_sgs = {
            sample['id']: sample['sequencingGroups']
            for participant in participants
            for sample in participant['samples']
        }
        for sample_id, sgs in sample_sgs.items():
            for sg in sgs:
                if sg['type'].lower() not in self.sequencing_types:
                    continue

                sg_sample_id_map[sg['id']] = sample_id
                for assay in sg['assays']:
                    reads = assay['meta'].get('reads')
                    if not reads:
                        logging.warning(
                            f'{self.dataset} :: SG {sg["id"]} assay {assay["id"]} has no reads field'
                        )
                        continue

                    if assay_sg_id_map.get(assay['id']):
                        raise ValueError(
                            f'{self.dataset} :: Assay {assay["id"]} has multiple SGs: {assay_sg_id_map[assay["id"]]} and {sg["id"]}'
                        )
                    assay_sg_id_map[assay['id']] = sg['id']

                    if isinstance(reads, dict):
                        assay_filepaths_filesizes[assay['id']].append(
                            (
                                reads.get('location'),
                                reads.get('size'),
                            )
                        )
                        continue

                    for read in reads:
                        if not isinstance(read, dict):
                            logging.error(
                                f'{self.dataset} :: Got {type(read)} read for SG {sg["id"]}, expected dict: {read}'
                            )
                            continue

                        assay_filepaths_filesizes[assay['id']].append(
                            (
                                read.get('location'),
                                read.get('size'),
                            )
                        )

        return sg_sample_id_map, assay_sg_id_map, assay_filepaths_filesizes

    async def get_analysis_cram_paths_for_dataset_sgs(
        self,
        assay_sg_id_map: dict[int, str],
    ) -> dict[str, dict[int, str]]:
        """
        Fetches all CRAMs for the list of sgs in the given dataset.
        Returns a dict mapping {sg_id : (analysis_id, cram_path) }
        """
        sg_ids = list(assay_sg_id_map.values())

        logging.getLogger().setLevel(logging.WARN)
        analyses_query_result = await query_async(
            QUERY_SG_ANALYSES,
            {'dataset': self.dataset, 'sgId': sg_ids, 'analysisTypes': ['CRAM']},
        )
        logging.getLogger().setLevel(logging.INFO)

        analyses = analyses_query_result['sequencingGroups']
        analyses = self.get_most_recent_analyses_by_sg(analyses_list=analyses)

        # Report any crams missing the sequencing type
        crams_with_missing_seq_type = [
            analysis['id']
            for analysis in analyses.values()
            if 'sequencing_type' not in analysis['meta']
        ]
        if crams_with_missing_seq_type:
            raise ValueError(
                f'{self.dataset} :: CRAM analyses are missing sequencing_type field: {crams_with_missing_seq_type}'
            )

        # For each sg id, collect the analysis id and cram paths
        sg_cram_paths: dict[str, dict[int, str]] = defaultdict(dict)
        for sg_id, analysis in analyses.items():
            cram_path = analysis['output']
            if not cram_path.startswith('gs://') or not cram_path.endswith('.cram'):
                logging.warning(
                    f'Analysis {analysis["id"]} invalid output path: {analysis["output"]}'
                )
                continue

            sg_cram_paths[sg_id][analysis['id']] = analysis['output']

        return sg_cram_paths

    async def analyses_for_sgs_without_crams(self, sgs_without_crams: list[str]):
        """Checks if other completed analyses exist for samples without completed crams"""

        all_sg_analyses: dict[str, list[dict[str, int | str]]] = defaultdict(list)

        logging.getLogger().setLevel(logging.WARN)
        sg_analyse_query_result = await query_async(
            QUERY_SG_ANALYSES,
            {
                'dataset': self.dataset,
                'sgIds': sgs_without_crams,
                'analysisTypes': [t for t in await get_analysis_types() if t != 'cram'],
            },
        )
        logging.getLogger().setLevel(logging.INFO)

        sg_analyses = sg_analyse_query_result['sequencingGroups']

        for sg_analysis in sg_analyses:
            sg_id = sg_analysis['id']

            for analysis in sg_analysis['analyses']:
                analysis_entry = {
                    'analysis_id': analysis['id'],
                    'analysis_type': analysis['type'],
                    'analysis_output': analysis['output'],
                    'timestamp_completed': analysis['timestampCompleted'],
                }
                all_sg_analyses[sg_id].append(analysis_entry)

        if all_sg_analyses:
            for sg_without_cram, completed_analyses in all_sg_analyses.items():
                for completed_analysis in completed_analyses:
                    logging.warning(
                        f'{self.dataset} :: SG {sg_without_cram} missing CRAM but has analysis {completed_analysis}'
                    )

    async def get_complete_and_incomplete_sgs(
        self,
        assay_sg_id_map: dict[int, str],
        sg_cram_paths: dict[str, dict[int, str]],
    ) -> dict[str, Any]:
        """
        Returns a dictionary containing two categories of sequencing groups:
         - the completed sgs which have finished aligning and have a cram, as a dict mapping
            the sg_id to the analysis_ids
         - the incomplete sgs where the alignment hasn't completed and no cram exists, as a list
        """
        # Get all the unique cram paths to check
        cram_paths = set()
        for analyses in sg_cram_paths.values():
            cram_paths.update(list(analyses.values()))

        # Check the analysis CRAM paths actually exist in the bucket
        buckets_subdirs = self.get_gcs_bucket_subdirs_to_search(list(cram_paths))
        crams_in_bucket = self.find_files_in_gcs_buckets_subdirs(
            buckets_subdirs,
            ('cram',),
        )

        # Incomplete SGs initialised as the SGs without a completed CRAM
        incomplete_sgs = set(assay_sg_id_map.values()).difference(
            set(sg_cram_paths.keys())
        )

        # Completed SGs have a CRAM file in the bucket that matches the path in Metamist
        completed_sgs = {}
        for sg_id, analysis in sg_cram_paths.items():
            for analysis_id, cram_path in analysis.items():
                if cram_path in crams_in_bucket:
                    completed_sgs[sg_id] = analysis_id
                    continue
                incomplete_sgs.update(sg_id)

        if incomplete_sgs:
            logging.warning(
                f'{self.dataset} :: {len(incomplete_sgs)} SGs without CRAMs found: {list(incomplete_sgs)}'
            )
            await self.analyses_for_sgs_without_crams(list(incomplete_sgs))

        return {'complete': completed_sgs, 'incomplete': list(incomplete_sgs)}

    async def check_for_uningested_or_moved_assays(  # pylint: disable=R0914
        self,
        bucket_name: str,
        assay_filepaths_filesizes: dict[int, list[tuple[str, int]]],
        completed_sgs: dict[str, list[int]],
        sg_sample_id_map: dict[str, str],
        assay_sg_id_map: dict[int, str],
        sample_internal_external_id_map: dict[str, str],
    ):
        """
        Compares the assays read files in a Metamist dataset to the read files found in the
        main-upload bucket.

        Input:  The upload bucket name, the {assay_id : (read_path, read_size)} mapping,
                the completed SGs, the { SG_ID : sample_ID } mapping, the { assay_id : SG_ID } mapping,
                and the { sample_id : sample_external_id } mapping.

        Returns: 1. Paths to assays that have not yet been ingested - checks if any completed
                    sample external IDs are in the filename and includes this in output.
                 2. A dict mapping assay IDs to GS filepaths that have been ingested,
                    but where the path in the bucket is different to the path for this
                    assay in Metamist. If the filenames and file sizes match,
                    these are identified as assay files that have been moved from
                    their original location to a new location.
                 3. The assay read files that have been deleted/moved.
        """
        # Get all the paths to assay data anywhere in the main-upload bucket
        assay_paths_in_bucket = self.find_assay_files_in_gcs_bucket(
            bucket_name, self.file_types
        )

        # Flatten all the Metamist assay file paths and sizes into a single list
        assay_paths_sizes_in_metamist: list[tuple[str, int]] = []
        for assay in assay_filepaths_filesizes.values():
            assay_paths_sizes_in_metamist.extend(assay)

        # Find the paths that exist in the bucket and not in metamist
        assay_paths_in_metamist = [
            path_size[0] for path_size in assay_paths_sizes_in_metamist
        ]
        uningested_assay_paths = set(assay_paths_in_bucket).difference(
            set(assay_paths_in_metamist)
        )
        # Find the paths that exist in metamist and not in the bucket
        metamist_paths_to_nowhere = set(assay_paths_in_metamist).difference(
            set(assay_paths_in_bucket)
        )

        # Strip the metamist paths into just filenames
        # Map each file name to its file size and path
        metamist_assay_file_size_map = {
            os.path.basename(path_size[0]): path_size[1]
            for path_size in assay_paths_sizes_in_metamist
        }
        metamist_assay_file_path_map = {
            os.path.basename(path_size[0]): path_size[0]
            for path_size in assay_paths_sizes_in_metamist
        }

        # Identify if any paths are to files that have actually just been moved
        # by checking if they are in the bucket but not metamist
        ingested_and_moved_filepaths = []
        new_assay_path_sizes = {}
        for path in uningested_assay_paths:
            filename = os.path.basename(path)
            # If the file in the bucket has the exact same name and size as one in metamist, assume its the same
            if filename in metamist_assay_file_size_map.keys():
                filesize = await self.file_size(path)
                if filesize == metamist_assay_file_size_map.get(filename):
                    ingested_and_moved_filepaths.append(
                        (path, metamist_assay_file_path_map.get(filename))
                    )
                    new_assay_path_sizes[path] = filesize
        logging.info(
            f'Found {len(ingested_and_moved_filepaths)} ingested files that have been moved'
        )

        # If the file has just been moved, we consider it ingested
        uningested_assay_paths -= {
            bucket_path for bucket_path, _ in ingested_and_moved_filepaths
        }

        # Check the list of uningested paths to see if any of them contain sample ids for ingested samples
        # This could happen when we ingest a fastq read pair for a sample, and additional read files were provided
        # but not ingested, such as bams and vcfs.
        uningested_reads: dict[str, list[tuple[str, str, str]]] = defaultdict(
            list, {k: [] for k in uningested_assay_paths}
        )
        for sg_id, analysis_ids in completed_sgs.items():
            try:
                sample_id = sg_sample_id_map[sg_id]
                sample_ext_id = sample_internal_external_id_map[sample_id]
                for uningested_assay in uningested_assay_paths:
                    if sample_ext_id not in uningested_assay:
                        continue
                    uningested_reads[uningested_assay].append(
                        (sg_id, sample_id, sample_ext_id)
                    )
            except KeyError:
                logging.warning(
                    f'{sg_id} from analyses: {analysis_ids} not found in SG-sample map.'
                )

        # flip the assay id : reads mapping to identify assays by their reads
        reads_assays = {}
        for assay_id, reads_sizes in assay_filepaths_filesizes.items():
            for read_size in reads_sizes:
                reads_assays[read_size[0]] = assay_id

        # Collect the assays for files that have been ingested and moved to a different bucket location
        assays_moved_paths = []
        for bucket_path, metamist_path in ingested_and_moved_filepaths:
            assay_id = reads_assays.get(metamist_path)
            sg_id = assay_sg_id_map.get(assay_id)
            analysis_id = completed_sgs.get(sg_id)
            if sg_id not in AuditHelper.EXCLUDED_SGS and analysis_id:
                filesize = new_assay_path_sizes[bucket_path]
                assays_moved_paths.append(
                    AssayReportEntry(
                        sg_id=sg_id,
                        assay_id=assay_id,
                        assay_file_path=bucket_path,
                        analysis_id=analysis_id,
                        filesize=filesize,
                    )
                )

        return uningested_reads, assays_moved_paths, metamist_paths_to_nowhere

    async def get_reads_to_delete_or_ingest(
        self,
        bucket_name: str,
        completed_sgs: dict[str, list[int]],
        assay_filepaths_filesizes: dict[int, list[tuple[str, int]]],
        sg_sample_id_map: dict[str, str],
        assay_sg_id_map: dict[int, str],
        sample_internal_external_id_map: dict[str, str],
    ) -> tuple[list, list]:
        """
        Inputs: 1. List of samples which have completed CRAMs
                2. Dictionary mapping of assay IDs to assay file paths and file sizes
                3. Dictionary mapping sample IDs to assay IDs
                4. Dictionary mapping assay IDs to sample IDs
        Returns a tuple of two lists, each containing assays IDs and assay file paths.
        The first containins reads which can be deleted, the second containing reads to ingest.
        The sample id, assay id, and analysis id (of completed cram) are included in the delete list.
        """
        # Check for uningested assay data that may be hiding or assay data that has been moved
        (
            reads_to_ingest,
            moved_assays_to_delete,
            metamist_paths_to_nowhere,
        ) = await self.check_for_uningested_or_moved_assays(
            bucket_name,
            assay_filepaths_filesizes,
            completed_sgs,
            sg_sample_id_map,
            assay_sg_id_map,
            sample_internal_external_id_map,
        )

        # Create a mapping of sg id: assay id - use defaultdict in case a sg has several assays
        sg_assays_id_map: dict[str, list[int]] = defaultdict(list)
        for assay_id, sg_id in assay_sg_id_map.items():
            sg_assays_id_map[sg_id].append(assay_id)

        assay_reads_to_delete = []
        for sg_id, analysis_id in completed_sgs.items():
            if sg_id in AuditHelper.EXCLUDED_SGS:
                continue
            assay_ids = sg_assays_id_map[sg_id]
            for assay_id in assay_ids:
                assay_read_paths = assay_filepaths_filesizes[assay_id]
                for path, size in assay_read_paths:
                    if path in metamist_paths_to_nowhere:
                        continue
                    filesize = size
                    assay_reads_to_delete.append(
                        AssayReportEntry(
                            sg_id=sg_id,
                            assay_id=assay_id,
                            assay_file_path=path,
                            analysis_id=analysis_id,
                            filesize=filesize,
                        )
                    )

        reads_to_delete = assay_reads_to_delete + moved_assays_to_delete

        return reads_to_delete, reads_to_ingest

    @staticmethod
    def find_crams_for_reads_to_ingest(
        reads_to_ingest: dict[str, list],
        sg_cram_paths: dict[str, dict[int, str]],
    ) -> list[tuple[str, str, str, str, int, str]]:
        """
        Compares the external sample IDs for samples with completed CRAMs against the
        uningested read files. This may turn up results for cases where multiple read types
        have been provided for a sample, but only one type was ingested and used for alignment.
        """
        possible_assay_ingests = []
        for assay_path, sample_tuples in reads_to_ingest.items():
            if not sample_tuples:
                # If no samples detected in filename, add the path in an empty tuple
                possible_assay_ingests.append((assay_path, '', '', '', 0, ''))
                continue
            for sample_tuple in sample_tuples:
                # Else get the completed CRAM analysis id and path for the sample
                sg_id, sample_id, sample_external_id = sample_tuple
                sg_cram = sg_cram_paths[sg_id]
                analysis_id = int(list(sg_cram.keys())[0])
                cram_path = sg_cram[analysis_id]
                possible_assay_ingests.append(
                    (
                        assay_path,
                        sg_id,
                        sample_id,
                        sample_external_id,
                        analysis_id,
                        cram_path,
                    )
                )

        return possible_assay_ingests



"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 7.4.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from metamist.api_client import ApiClient, Endpoint as _Endpoint
from metamist.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types,
    async_wrap
)
from metamist.model.http_validation_error import HTTPValidationError


class ImportApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        self.import_individual_metadata_manifest_async = async_wrap(self.import_individual_metadata_manifest)
        self.import_individual_metadata_manifest_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/import/{project}/individual-metadata-manifest',
                'operation_id': 'import_individual_metadata_manifest',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'file',
                    'delimiter',
                    'extra_participants_method',
                ],
                'required': [
                    'project',
                    'file',
                ],
                'nullable': [
                    'delimiter',
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'file':
                        (file_type,),
                    'delimiter':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                    'extra_participants_method':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                },
                'attribute_map': {
                    'project': 'project',
                    'file': 'file',
                    'delimiter': 'delimiter',
                    'extra_participants_method': 'extra_participants_method',
                },
                'location_map': {
                    'project': 'path',
                    'file': 'form',
                    'delimiter': 'query',
                    'extra_participants_method': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'multipart/form-data'
                ]
            },
            api_client=api_client
        )

    def import_individual_metadata_manifest(
        self,
        project,
        file,
        **kwargs
    ):
        """Import Individual Metadata Manifest  # noqa: E501

        Import individual metadata manifest  :param extra_participants_method: If extra participants are in the uploaded file,     add a PARTICIPANT entry for them  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.import_individual_metadata_manifest(project, file, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            file (file_type):

        Keyword Args:
            delimiter (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            extra_participants_method (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['file'] = \
            file
        return self.import_individual_metadata_manifest_endpoint.call_with_http_info(**kwargs)




from test.testbase import DbIsolatedTest, run_as_sync

from pymysql.err import IntegrityError

from db.python.layers import FamilyLayer, ParticipantLayer, SampleLayer
from db.python.utils import NotFoundError
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)


class TestParticipant(DbIsolatedTest):
    """Test participant external ids"""

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.player = ParticipantLayer(self.connection)

        self.p1 = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'P1',
                    'CONTROL': '86',
                    'KAOS': 'shoe',
                },
            )
        )
        self.p1_external_ids = {k.lower(): v for k, v in self.p1.external_ids.items()}

        self.p2 = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'P2'},
            )
        )

    async def insert(self, participant_id, org, external_id):
        """Directly insert into participant_external_id table"""
        query = """
        INSERT INTO participant_external_id (project, participant_id, name, external_id, audit_log_id)
        VALUES (:project, :id, :org, :external_id, :audit_log_id)
        """
        values = {
            'project': self.project_id,
            'id': participant_id,
            'org': org,
            'external_id': external_id,
            'audit_log_id': await self.audit_log_id(),
        }
        await self.connection.connection.execute(query, values)

    @run_as_sync
    async def test_constraints(self):
        """Verify that database constraints prevent duplicate external_ids"""
        # Can't have two primary eids
        with self.assertRaises(IntegrityError):
            await self.insert(self.p1.id, PRIMARY_EXTERNAL_ORG, 'P86')

        # Can't have two eids in the same external organisation
        with self.assertRaises(IntegrityError):
            await self.insert(self.p1.id, 'CONTROL', 'Maxwell')

        # Can have eids in lots of organisations, but not if the eid duplicates one in a different org
        await self.insert(self.p1.id, 'OTHER1', 'Maxwell')
        with self.assertRaises(IntegrityError):
            await self.insert(self.p1.id, 'OTHER2', '86')

        # Another participant can't have the same eid
        with self.assertRaises(IntegrityError):
            await self.insert(self.p2.id, 'CONTROL', '86')

        # But it can have its own eid
        await self.insert(self.p2.id, 'CONTROL', '99')

    @run_as_sync
    async def test_insert(self):
        """Test inserting new participants with various external_id combinations"""
        with self.assertRaises(ValueError):
            _ = await self.player.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: None},
                )
            )

        with self.assertRaises(ValueError):
            _ = await self.player.upsert_participant(
                ParticipantUpsertInternal(external_ids={'OTHER': 'P1'})
            )

        result = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'P10',
                    'A': 'A1',
                    'B': None,
                    'C': 'C1',
                },
            )
        )
        participants = await self.player.get_participants_by_ids([result.id])
        self.assertDictEqual(
            participants[0].external_ids,
            {PRIMARY_EXTERNAL_ORG: 'P10', 'a': 'A1', 'c': 'C1'},
        )

    @run_as_sync
    async def test_update(self):
        """Test updating existing participants with various external_id combinations"""
        with self.assertRaises(ValueError):
            _ = await self.player.upsert_participant(
                ParticipantUpsertInternal(
                    id=self.p1.id,
                    external_ids={PRIMARY_EXTERNAL_ORG: None},
                )
            )

        result = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                id=self.p1.id,
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'P1B',
                    'CONTROL': '86B',
                    'KAOS': None,
                    'B': None,
                    'C': 'A1',
                },
            )
        )
        participants = await self.player.get_participants_by_ids([result.id])
        self.assertDictEqual(
            participants[0].external_ids,
            {PRIMARY_EXTERNAL_ORG: 'P1B', 'control': '86B', 'c': 'A1'},
        )

    @run_as_sync
    async def test_fill_in_missing(self):
        """Exercise fill_in_missing_participants() method"""
        slayer = SampleLayer(self.connection)

        s1 = await slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'E1'}, participant_id=self.p1.id
            ),
        )
        sa = await slayer.upsert_sample(
            SampleUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'EA', 'foo': 'FA'})
        )
        sb = await slayer.upsert_sample(
            SampleUpsertInternal(external_ids={PRIMARY_EXTERNAL_ORG: 'EB', 'foo': 'FB'})
        )

        result = await self.player.fill_in_missing_participants()
        self.assertEqual(result, 'Updated 2 records')

        samples = await slayer.get_samples_by(sample_ids=[s1.id, sa.id, sb.id])

        participants = await self.player.get_participants_by_ids(
            [s.participant_id for s in samples]
        )
        p_map = {p.id: p for p in participants}

        for s in samples:
            expected_eids = self.p1_external_ids if s.id == s1.id else s.external_ids
            self.assertEqual(p_map[s.participant_id].external_ids, expected_eids)

    @run_as_sync
    async def test_get_by_families(self):
        """Exercise get_participants_by_families() method"""
        flayer = FamilyLayer(self.connection)
        fid = await flayer.create_family(external_id='Jones')

        child = await self.player.upsert_participant(
            ParticipantUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'P20', 'd': 'D20'}
            ),
        )

        await self.player.add_participant_to_family(
            family_id=fid,
            participant_id=child.id,
            paternal_id=self.p1.id,
            maternal_id=self.p2.id,
            affected=2,
        )

        result = await self.player.get_participants_by_families([fid])
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[fid]), 1)
        self.assertDictEqual(
            result[fid][0].external_ids, {PRIMARY_EXTERNAL_ORG: 'P20', 'd': 'D20'}
        )

    @run_as_sync
    async def test_update_many(self):
        """Exercise update_many_participant_external_ids() method"""
        result = await self.player.update_many_participant_external_ids(
            {self.p1.id: 'P1B', self.p2.id: 'P2B'}
        )
        self.assertTrue(result)

        participants = await self.player.get_participants_by_ids(
            [self.p1.id, self.p2.id]
        )
        p_map = {p.id: p for p in participants}
        outp1 = p_map[self.p1.id]
        outp2 = p_map[self.p2.id]
        self.assertDictEqual(
            outp1.external_ids,
            {PRIMARY_EXTERNAL_ORG: 'P1B', 'control': '86', 'kaos': 'shoe'},
        )
        self.assertDictEqual(outp2.external_ids, {PRIMARY_EXTERNAL_ORG: 'P2B'})

    @run_as_sync
    async def test_get_etoi_map(self):
        """Exercise get_external_participant_id_to_internal_sequencing_group_id_map() method"""
        slayer = SampleLayer(self.connection)

        _ = await slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SE1'}, participant_id=self.p1.id
            ),
        )

        s2 = await slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'SE2', 'other': 'SO1'},
                participant_id=self.p1.id,
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[],
                    ),
                ],
            ),
        )

        result = await self.player.get_external_participant_id_to_internal_sequencing_group_id_map(
            self.project_id
        )
        self.assertEqual(len(result), 3)
        for eid, sgid in result:
            self.assertIn(eid, self.p1.external_ids.values())
            self.assertEqual(sgid, s2.sequencing_groups[0].id)


class TestSample(DbIsolatedTest):
    """Test sample external ids"""

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.slayer = SampleLayer(self.connection)

        self.s1 = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'S1',
                    'CONTROL': '86',
                    'KAOS': 'shoe',
                },
            )
        )

        self.s2 = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'S2'},
            )
        )

    async def insert(self, sample_id, org, external_id):
        """Directly insert into sample_external_id table"""
        query = """
        INSERT INTO sample_external_id (project, sample_id, name, external_id, audit_log_id)
        VALUES (:project, :id, :org, :external_id, :audit_log_id)
        """
        values = {
            'project': self.project_id,
            'id': sample_id,
            'org': org,
            'external_id': external_id,
            'audit_log_id': await self.audit_log_id(),
        }
        await self.connection.connection.execute(query, values)

    @run_as_sync
    async def test_constraints(self):
        """Verify that database constraints prevent duplicate external_ids"""
        # Can't have two primary eids
        with self.assertRaises(IntegrityError):
            await self.insert(self.s1.id, PRIMARY_EXTERNAL_ORG, 'S86')

        # Can't have two eids in the same external organisation
        with self.assertRaises(IntegrityError):
            await self.insert(self.s1.id, 'CONTROL', 'Maxwell')

        # Can have eids in lots of organisations, but not if the eid duplicates one in a different org
        await self.insert(self.s1.id, 'OTHER1', 'Maxwell')
        with self.assertRaises(IntegrityError):
            await self.insert(self.s1.id, 'OTHER2', '86')

        # Another sample can't have the same eid
        with self.assertRaises(IntegrityError):
            await self.insert(self.s2.id, 'CONTROL', '86')

        # But it can have its own eid
        await self.insert(self.s2.id, 'CONTROL', '99')

    @run_as_sync
    async def test_insert(self):
        """Test inserting new samples with various external_id combinations"""
        with self.assertRaises(ValueError):
            _ = await self.slayer.upsert_sample(
                SampleUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: None},
                )
            )

        with self.assertRaises(ValueError):
            _ = await self.slayer.upsert_sample(
                SampleUpsertInternal(external_ids={'OTHER': 'S1'})
            )

        result = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'S10',
                    'A': 'A1',
                    'B': None,
                    'C': 'C1',
                },
            )
        )
        sample = await self.slayer.get_sample_by_id(result.id)
        self.assertDictEqual(
            sample.external_ids, {PRIMARY_EXTERNAL_ORG: 'S10', 'a': 'A1', 'c': 'C1'}
        )

    @run_as_sync
    async def test_update(self):
        """Test updating existing samples with various external_id combinations"""
        with self.assertRaises(ValueError):
            _ = await self.slayer.upsert_sample(
                SampleUpsertInternal(
                    id=self.s1.id,
                    external_ids={PRIMARY_EXTERNAL_ORG: None},
                )
            )

        result = await self.slayer.upsert_sample(
            SampleUpsertInternal(
                id=self.s1.id,
                external_ids={
                    PRIMARY_EXTERNAL_ORG: 'S1B',
                    'CONTROL': '86B',
                    'KAOS': None,
                    'B': None,
                    'C': 'A1',
                },
            )
        )
        sample = await self.slayer.get_sample_by_id(result.id)
        self.assertDictEqual(
            sample.external_ids,
            {PRIMARY_EXTERNAL_ORG: 'S1B', 'control': '86B', 'c': 'A1'},
        )

    @run_as_sync
    async def test_get_single(self):
        """Exercise get_single_by_external_id() method"""
        with self.assertRaises(NotFoundError):
            _ = await self.slayer.get_single_by_external_id(
                'non-existent', self.project_id
            )

        result = await self.slayer.get_single_by_external_id('86', self.project_id)
        self.assertEqual(result.id, self.s1.id)

        result = await self.slayer.get_single_by_external_id('S2', self.project_id)
        self.assertEqual(result.id, self.s2.id)

    @run_as_sync
    async def test_get_internal_to_external(self):
        """Exercise get_internal_to_external_sample_id_map() method"""
        result = await self.slayer.get_internal_to_external_sample_id_map(
            [self.s1.id, self.s2.id]
        )
        self.assertDictEqual(result, {self.s1.id: 'S1', self.s2.id: 'S2'})

    @run_as_sync
    async def test_get_all(self):
        """Exercise get_all_sample_id_map_by_internal_ids() method"""
        result = await self.slayer.get_all_sample_id_map_by_internal_ids(
            self.project_id
        )
        self.assertDictEqual(result, {self.s1.id: 'S1', self.s2.id: 'S2'})

    @run_as_sync
    async def test_get_history(self):
        """Exercise get_history_of_sample() method"""
        # First create some history
        await self.slayer.upsert_sample(
            SampleUpsertInternal(id=self.s1.id, meta={'foo': 'bar'})
        )

        await self.slayer.upsert_sample(
            SampleUpsertInternal(
                id=self.s1.id,
                external_ids={'fruit': 'apple'},
                meta={'fruit': 'banana'},
            )
        )

        sample = await self.slayer.get_sample_by_id(self.s1.id)

        result = await self.slayer.get_history_of_sample(self.s1.id)
        self.assertEqual(len(result), 3)
        self.assertDictEqual(result[0].meta, {})
        self.assertDictEqual(result[1].meta, {'foo': 'bar'})
        self.assertDictEqual(result[2].meta, {'foo': 'bar', 'fruit': 'banana'})
        self.assertDictEqual(result[2].meta, sample.meta)



# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from metamist.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from metamist.model.analysis import Analysis
from metamist.model.analysis_cost_record import AnalysisCostRecord
from metamist.model.analysis_cost_record_batch import AnalysisCostRecordBatch
from metamist.model.analysis_cost_record_batch_job import AnalysisCostRecordBatchJob
from metamist.model.analysis_cost_record_category import AnalysisCostRecordCategory
from metamist.model.analysis_cost_record_cromwell_subworkflow import AnalysisCostRecordCromwellSubworkflow
from metamist.model.analysis_cost_record_cromwell_workflow import AnalysisCostRecordCromwellWorkflow
from metamist.model.analysis_cost_record_seq_group import AnalysisCostRecordSeqGroup
from metamist.model.analysis_cost_record_sku import AnalysisCostRecordSku
from metamist.model.analysis_cost_record_topic import AnalysisCostRecordTopic
from metamist.model.analysis_cost_record_total import AnalysisCostRecordTotal
from metamist.model.analysis_cost_record_wdl_task import AnalysisCostRecordWdlTask
from metamist.model.analysis_query_model import AnalysisQueryModel
from metamist.model.analysis_runner import AnalysisRunner
from metamist.model.analysis_stats import AnalysisStats
from metamist.model.analysis_status import AnalysisStatus
from metamist.model.analysis_update_model import AnalysisUpdateModel
from metamist.model.assay import Assay
from metamist.model.assay_query_criteria import AssayQueryCriteria
from metamist.model.assay_upsert import AssayUpsert
from metamist.model.billing_column import BillingColumn
from metamist.model.billing_cost_budget_record import BillingCostBudgetRecord
from metamist.model.billing_cost_details_record import BillingCostDetailsRecord
from metamist.model.billing_source import BillingSource
from metamist.model.billing_time_column import BillingTimeColumn
from metamist.model.billing_time_periods import BillingTimePeriods
from metamist.model.billing_total_cost_query_model import BillingTotalCostQueryModel
from metamist.model.billing_total_cost_record import BillingTotalCostRecord
from metamist.model.body_create_cohort_from_criteria import BodyCreateCohortFromCriteria
from metamist.model.body_export_project_participants import BodyExportProjectParticipants
from metamist.model.body_get_latest_complete_analysis_for_type_post import BodyGetLatestCompleteAnalysisForTypePost
from metamist.model.body_get_project_insights_details import BodyGetProjectInsightsDetails
from metamist.model.body_get_project_insights_summary import BodyGetProjectInsightsSummary
from metamist.model.body_get_proportionate_map import BodyGetProportionateMap
from metamist.model.cohort_body import CohortBody
from metamist.model.cohort_criteria import CohortCriteria
from metamist.model.cohort_template import CohortTemplate
from metamist.model.error_response import ErrorResponse
from metamist.model.export_project_participant_fields import ExportProjectParticipantFields
from metamist.model.export_type import ExportType
from metamist.model.extra_participant_importer_handler import ExtraParticipantImporterHandler
from metamist.model.family import Family
from metamist.model.family_search_response_data import FamilySearchResponseData
from metamist.model.family_simple import FamilySimple
from metamist.model.family_update_model import FamilyUpdateModel
from metamist.model.generic_filter_any import GenericFilterAny
from metamist.model.generic_filter_int import GenericFilterInt
from metamist.model.generic_filter_str import GenericFilterStr
from metamist.model.get_samples_criteria import GetSamplesCriteria
from metamist.model.http_validation_error import HTTPValidationError
from metamist.model.meta_search_entity_prefix import MetaSearchEntityPrefix
from metamist.model.nested_participant import NestedParticipant
from metamist.model.nested_sample import NestedSample
from metamist.model.nested_sequencing_group import NestedSequencingGroup
from metamist.model.new_cohort import NewCohort
from metamist.model.participant_grid_assay_filter import ParticipantGridAssayFilter
from metamist.model.participant_grid_family_filter import ParticipantGridFamilyFilter
from metamist.model.participant_grid_participant_filter import ParticipantGridParticipantFilter
from metamist.model.participant_grid_sample_filter import ParticipantGridSampleFilter
from metamist.model.participant_grid_sequencing_group_filter import ParticipantGridSequencingGroupFilter
from metamist.model.participant_search_response_data import ParticipantSearchResponseData
from metamist.model.participant_upsert import ParticipantUpsert
from metamist.model.project import Project
from metamist.model.project_insights_details import ProjectInsightsDetails
from metamist.model.project_insights_summary import ProjectInsightsSummary
from metamist.model.project_member_role import ProjectMemberRole
from metamist.model.project_member_update import ProjectMemberUpdate
from metamist.model.project_participant_grid_field import ProjectParticipantGridField
from metamist.model.project_participant_grid_filter import ProjectParticipantGridFilter
from metamist.model.project_participant_grid_filter_type import ProjectParticipantGridFilterType
from metamist.model.project_participant_grid_response import ProjectParticipantGridResponse
from metamist.model.project_summary import ProjectSummary
from metamist.model.proportional_date_temporal_method import ProportionalDateTemporalMethod
from metamist.model.query_participant_criteria import QueryParticipantCriteria
from metamist.model.sample_search_response_data import SampleSearchResponseData
from metamist.model.sample_upsert import SampleUpsert
from metamist.model.search_response import SearchResponse
from metamist.model.search_response_model import SearchResponseModel
from metamist.model.search_response_type import SearchResponseType
from metamist.model.seqr_dataset_type import SeqrDatasetType
from metamist.model.sequencing_group_meta_update_model import SequencingGroupMetaUpdateModel
from metamist.model.sequencing_group_search_response_data import SequencingGroupSearchResponseData
from metamist.model.sequencing_group_upsert import SequencingGroupUpsert
from metamist.model.validation_error import ValidationError
from metamist.model.web_project import WebProject



import datetime
from test.testbase import DbIsolatedTest, run_as_sync

from pymysql.err import IntegrityError

from db.python.filters import GenericFilter
from db.python.layers import CohortLayer, SampleLayer
from db.python.tables.cohort import CohortFilter
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
)
from models.models.cohort import (
    CohortCriteria,
    CohortCriteriaInternal,
    CohortTemplate,
    CohortTemplateInternal,
    NewCohort,
    NewCohortInternal,
)
from models.utils.cohort_id_format import cohort_id_format
from models.utils.sequencing_group_id_format import sequencing_group_id_format


class TestCohortBasic(DbIsolatedTest):
    """Test custom cohort endpoints"""

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.cohortl = CohortLayer(self.connection)

    @run_as_sync
    async def test_create_cohort_missing_args(self):
        """Can't create cohort with neither criteria nor template"""
        with self.assertRaises(ValueError):
            _ = await self.cohortl.create_cohort_from_criteria(
                project_to_write=self.project_id,
                description='No criteria or template',
                cohort_name='Broken cohort',
                dry_run=False,
            )

    # These tests are disabled because the move to an Internal Model means that verification happens in the route not the layer
    # @run_as_sync
    # async def test_create_cohort_bad_project(self):
    #     """Can't create cohort in invalid project"""
    #     with self.assertRaises((Forbidden, NotFoundError)):
    #         _ = await self.cohortl.create_cohort_from_criteria(
    #             project_to_write=self.project_id,
    #             description='Cohort based on a missing project',
    #             cohort_name='Bad-project cohort',
    #             dry_run=False,
    #             cohort_criteria=CohortCriteriaInternal(projects=[5]),
    #         )

    # @run_as_sync
    # async def test_create_template_bad_project(self):
    #     """Can't create template in invalid project"""
    #     with self.assertRaises((Forbidden, NotFoundError)):
    #         _ = await self.cohortl.create_cohort_template(
    #             project=self.project_id,
    #             cohort_template=CohortTemplate(
    #                 id=None,
    #                 name='Bad-project template',
    #                 description='Template based on a missing project',
    #                 criteria=CohortCriteria(projects=['nonexistent']),
    #             ),
    #         )

    @run_as_sync
    async def test_create_empty_cohort(self):
        """Create cohort from empty criteria"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with no entries',
            cohort_name='Empty cohort',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(projects=[self.project_id]),
        )
        self.assertIsInstance(result, NewCohortInternal)
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual([], result.sequencing_group_ids)

    @run_as_sync
    async def test_create_duplicate_cohort(self):
        """Can't create cohorts with duplicate names"""
        _ = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with no entries',
            cohort_name='Trial duplicate cohort',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(projects=[self.project_id]),
        )

        _ = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with no entries',
            cohort_name='Trial duplicate cohort',
            dry_run=True,
            cohort_criteria=CohortCriteriaInternal(projects=[self.project_id]),
        )

        with self.assertRaises(IntegrityError):
            _ = await self.cohortl.create_cohort_from_criteria(
                project_to_write=self.project_id,
                description='Cohort with no entries',
                cohort_name='Trial duplicate cohort',
                dry_run=False,
                cohort_criteria=CohortCriteriaInternal(projects=[self.project_id]),
            )

    @run_as_sync
    async def test_create_template_then_cohorts(self):
        """Test with template and cohort IDs out of sync, and creating from template"""
        tid = await self.cohortl.create_cohort_template(
            project=self.project_id,
            cohort_template=CohortTemplateInternal(
                id=None,
                name='Empty template',
                description='Template with no entries',
                criteria=CohortCriteriaInternal(projects=[self.project_id]),
                project=self.project_id,
            ),
        )

        _ = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with no entries',
            cohort_name='Another empty cohort',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(projects=[self.project_id]),
        )

        _ = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort from template',
            cohort_name='Cohort from empty template',
            dry_run=False,
            template_id=tid,
        )


class TestCohortQueries(DbIsolatedTest):
    """Test query-related custom cohort layer functions"""

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.cohortl = CohortLayer(self.connection)

    @run_as_sync
    async def test_id_query(self):
        """Exercise querying id against an empty database"""
        result = await self.cohortl.query(CohortFilter(id=GenericFilter(eq=42)))
        self.assertEqual([], result)

    @run_as_sync
    async def test_name_query(self):
        """Exercise querying name against an empty database"""
        result = await self.cohortl.query(
            CohortFilter(name=GenericFilter(eq='Unknown cohort'))
        )
        self.assertEqual([], result)

    @run_as_sync
    async def test_author_query(self):
        """Exercise querying author against an empty database"""
        result = await self.cohortl.query(
            CohortFilter(author=GenericFilter(eq='Alan Smithee'))
        )
        self.assertEqual([], result)

    @run_as_sync
    async def test_template_id_query(self):
        """Exercise querying template_id against an empty database"""
        result = await self.cohortl.query(
            CohortFilter(template_id=GenericFilter(eq=28))
        )
        self.assertEqual([], result)

    @run_as_sync
    async def test_timestamp_query(self):
        """Exercise querying timestamp against an empty database"""
        new_years_day = datetime.datetime(2024, 1, 1)
        result = await self.cohortl.query(
            CohortFilter(timestamp=GenericFilter(eq=new_years_day))
        )
        self.assertEqual([], result)

    @run_as_sync
    async def test_project_query(self):
        """Exercise querying project against an empty database"""
        result = await self.cohortl.query(CohortFilter(project=GenericFilter(eq=37)))
        self.assertEqual([], result)


def get_sample_model(
    eid, s_type='blood', sg_type='genome', tech='short-read', plat='illumina'
):
    """Create a minimal sample"""
    return SampleUpsertInternal(
        meta={},
        external_ids={PRIMARY_EXTERNAL_ORG: f'EXID{eid}'},
        type=s_type,
        sequencing_groups=[
            SequencingGroupUpsertInternal(
                type=sg_type,
                technology=tech,
                platform=plat,
                meta={},
                assays=[],
            ),
        ],
    )


class TestCohortData(DbIsolatedTest):
    """Test custom cohort endpoints that need some sequencing groups already set up"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self):
        super().setUp()
        self.cohortl = CohortLayer(self.connection)
        self.samplel = SampleLayer(self.connection)

        self.sA = await self.samplel.upsert_sample(get_sample_model('A'))
        self.sB = await self.samplel.upsert_sample(get_sample_model('B'))
        self.sC = await self.samplel.upsert_sample(
            get_sample_model('C', 'saliva', 'exome', 'long-read', 'ONT')
        )

        self.sgA = sequencing_group_id_format(self.sA.sequencing_groups[0].id)
        self.sgA_raw = self.sA.sequencing_groups[0].id
        self.sgB = sequencing_group_id_format(self.sB.sequencing_groups[0].id)
        self.sgB_raw = self.sB.sequencing_groups[0].id
        self.sgC = sequencing_group_id_format(self.sC.sequencing_groups[0].id)
        self.sgC_raw = self.sC.sequencing_groups[0].id

    @run_as_sync
    async def test_internal_external(self):
        """Test to_internal() and to_external() methods"""
        cc_external_dict = {
            'projects': [self.project_name],
            'sg_ids_internal': [self.sgB, self.sgC],
            'excluded_sgs_internal': [self.sgA],
            'sg_technology': ['short-read'],
            'sg_platform': ['illumina'],
            'sg_type': ['genome'],
            'sample_type': ['blood'],
        }

        cc_internal_dict = {
            'projects': [self.project_id],
            'sg_ids_internal_raw': [self.sgB_raw, self.sgC_raw],
            'excluded_sgs_internal_raw': [self.sgA_raw],
            'sg_technology': ['short-read'],
            'sg_platform': ['illumina'],
            'sg_type': ['genome'],
            'sample_type': ['blood'],
        }

        cc_external = CohortCriteria(**cc_external_dict)
        cc_internal = cc_external.to_internal(projects_internal=[self.project_id])
        self.assertIsInstance(cc_internal, CohortCriteriaInternal)
        self.assertDictEqual(cc_internal.model_dump(), cc_internal_dict)

        cc_ext_trip = cc_internal.to_external(project_names=[self.project_name])
        self.assertIsInstance(cc_ext_trip, CohortCriteria)
        self.assertDictEqual(cc_ext_trip.model_dump(), cc_external_dict)

        ctpl_internal_dict = {
            'id': 496,
            'name': 'My template',
            'description': 'Testing template',
            'criteria': cc_internal_dict,
            'project': self.project_id,
        }

        ctpl_internal = CohortTemplate(
            id=496,
            name='My template',
            description='Testing template',
            criteria=cc_external,
        ).to_internal(
            criteria_projects=[self.project_id], template_project=self.project_id
        )
        self.assertIsInstance(ctpl_internal, CohortTemplateInternal)
        self.assertDictEqual(ctpl_internal.model_dump(), ctpl_internal_dict)

    @run_as_sync
    async def test_create_cohort_by_sgs(self):
        """Create cohort by selecting sequencing groups"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with 1 SG',
            cohort_name='SG cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_ids_internal_raw=[self.sgB_raw],
            ),
        )
        self.assertIsInstance(result, NewCohortInternal)
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual([self.sgB_raw], result.sequencing_group_ids)

        external = result.to_external()
        self.assertIsInstance(external, NewCohort)
        self.assertIsInstance(external.cohort_id, str)
        self.assertEqual(external.cohort_id, cohort_id_format(result.cohort_id))
        self.assertEqual([self.sgB], external.sequencing_group_ids)
        self.assertEqual(False, external.dry_run)

    @run_as_sync
    async def test_create_cohort_by_excluded_sgs(self):
        """Create cohort by excluding sequencing groups"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort without 1 SG',
            cohort_name='SG cohort 2',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                excluded_sgs_internal_raw=[self.sgA_raw],
            ),
        )
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual(2, len(result.sequencing_group_ids))
        self.assertIn(self.sgB_raw, result.sequencing_group_ids)
        self.assertIn(self.sgC_raw, result.sequencing_group_ids)

    @run_as_sync
    async def test_create_cohort_by_technology(self):
        """Create cohort by selecting a technology"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Short-read cohort',
            cohort_name='Tech cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_technology=['short-read'],
            ),
        )
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual(2, len(result.sequencing_group_ids))
        self.assertIn(self.sgA_raw, result.sequencing_group_ids)
        self.assertIn(self.sgB_raw, result.sequencing_group_ids)

    @run_as_sync
    async def test_create_cohort_by_platform(self):
        """Create cohort by selecting a platform"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='ONT cohort',
            cohort_name='Platform cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_platform=['ONT'],
            ),
        )
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual([self.sgC_raw], result.sequencing_group_ids)

    @run_as_sync
    async def test_create_cohort_by_type(self):
        """Create cohort by selecting types"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Genome cohort',
            cohort_name='Type cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_type=['genome'],
            ),
        )
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual(2, len(result.sequencing_group_ids))
        self.assertIn(self.sgA_raw, result.sequencing_group_ids)
        self.assertIn(self.sgB_raw, result.sequencing_group_ids)

    @run_as_sync
    async def test_create_cohort_by_sample_type(self):
        """Create cohort by selecting sample types"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Sample cohort',
            cohort_name='Sample cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sample_type=['saliva'],
            ),
        )
        self.assertIsInstance(result.cohort_id, int)
        self.assertEqual([self.sgC_raw], result.sequencing_group_ids)

    @run_as_sync
    async def test_create_cohort_by_everything(self):
        """Create cohort by selecting a variety of fields"""
        result = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Everything cohort',
            cohort_name='Everything cohort 1',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_ids_internal_raw=[self.sgB_raw, self.sgC_raw],
                excluded_sgs_internal_raw=[self.sgA_raw],
                sg_technology=['short-read'],
                sg_platform=['illumina'],
                sg_type=['genome'],
                sample_type=['blood'],
            ),
        )
        self.assertEqual(1, len(result.sequencing_group_ids))
        self.assertIn(self.sgB_raw, result.sequencing_group_ids)

    @run_as_sync
    async def test_reevaluate_cohort(self):
        """Add another sample, then reevaluate a cohort template"""
        template = await self.cohortl.create_cohort_template(
            project=self.project_id,
            cohort_template=CohortTemplateInternal(
                id=None,
                name='Blood template',
                description='Template selecting blood',
                criteria=CohortCriteriaInternal(
                    projects=[self.project_id],
                    sample_type=['blood'],
                ),
                project=self.project_id,
            ),
        )

        coh1 = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Blood cohort',
            cohort_name='Blood cohort 1',
            dry_run=False,
            template_id=template,
        )
        self.assertEqual(2, len(coh1.sequencing_group_ids))

        sD = await self.samplel.upsert_sample(get_sample_model('D'))
        sgD_raw = sD.sequencing_groups[0].id

        coh2 = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Blood cohort',
            cohort_name='Blood cohort 2',
            dry_run=False,
            template_id=template,
        )
        self.assertEqual(3, len(coh2.sequencing_group_ids))

        self.assertNotIn(sgD_raw, coh1.sequencing_group_ids)
        self.assertIn(sgD_raw, coh2.sequencing_group_ids)

    @run_as_sync
    async def test_query_cohort(self):
        """Create a cohort and test that it is populated when queried"""
        created = await self.cohortl.create_cohort_from_criteria(
            project_to_write=self.project_id,
            description='Cohort with two samples',
            cohort_name='Duo cohort',
            dry_run=False,
            cohort_criteria=CohortCriteriaInternal(
                projects=[self.project_id],
                sg_ids_internal_raw=[self.sgA_raw, self.sgB_raw],
            ),
        )
        self.assertEqual(2, len(created.sequencing_group_ids))

        queried = await self.cohortl.query(
            CohortFilter(name=GenericFilter(eq='Duo cohort'))
        )
        self.assertEqual(1, len(queried))

        result = await self.cohortl.get_cohort_sequencing_group_ids(int(queried[0].id))
        self.assertEqual(2, len(result))
        self.assertIn(self.sA.sequencing_groups[0].id, result)
        self.assertIn(self.sB.sequencing_groups[0].id, result)



"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 7.4.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from metamist.api_client import ApiClient, Endpoint as _Endpoint
from metamist.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types,
    async_wrap
)
from metamist.model.analysis_cost_record import AnalysisCostRecord
from metamist.model.billing_column import BillingColumn
from metamist.model.billing_cost_budget_record import BillingCostBudgetRecord
from metamist.model.billing_total_cost_query_model import BillingTotalCostQueryModel
from metamist.model.billing_total_cost_record import BillingTotalCostRecord
from metamist.model.http_validation_error import HTTPValidationError


class BillingApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        self.cost_by_ar_guid_async = async_wrap(self.cost_by_ar_guid)
        self.cost_by_ar_guid_endpoint = _Endpoint(
            settings={
                'response_type': ([AnalysisCostRecord],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/cost-by-ar-guid/{ar_guid}',
                'operation_id': 'cost_by_ar_guid',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'ar_guid',
                ],
                'required': [
                    'ar_guid',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'ar_guid':
                        (str,),
                },
                'attribute_map': {
                    'ar_guid': 'ar_guid',
                },
                'location_map': {
                    'ar_guid': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.cost_by_batch_id_async = async_wrap(self.cost_by_batch_id)
        self.cost_by_batch_id_endpoint = _Endpoint(
            settings={
                'response_type': ([AnalysisCostRecord],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/cost-by-batch-id/{batch_id}',
                'operation_id': 'cost_by_batch_id',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'batch_id',
                ],
                'required': [
                    'batch_id',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'batch_id':
                        (str,),
                },
                'attribute_map': {
                    'batch_id': 'batch_id',
                },
                'location_map': {
                    'batch_id': 'path',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_compute_categories_async = async_wrap(self.get_compute_categories)
        self.get_compute_categories_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/compute-categories',
                'operation_id': 'get_compute_categories',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_cost_categories_async = async_wrap(self.get_cost_categories)
        self.get_cost_categories_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/categories',
                'operation_id': 'get_cost_categories',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_cromwell_sub_workflow_names_async = async_wrap(self.get_cromwell_sub_workflow_names)
        self.get_cromwell_sub_workflow_names_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/cromwell-sub-workflow-names',
                'operation_id': 'get_cromwell_sub_workflow_names',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_datasets_async = async_wrap(self.get_datasets)
        self.get_datasets_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/datasets',
                'operation_id': 'get_datasets',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_gcp_projects_async = async_wrap(self.get_gcp_projects)
        self.get_gcp_projects_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/gcp-projects',
                'operation_id': 'get_gcp_projects',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_invoice_months_async = async_wrap(self.get_invoice_months)
        self.get_invoice_months_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/invoice-months',
                'operation_id': 'get_invoice_months',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_namespaces_async = async_wrap(self.get_namespaces)
        self.get_namespaces_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/namespaces',
                'operation_id': 'get_namespaces',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_running_cost_async = async_wrap(self.get_running_cost)
        self.get_running_cost_endpoint = _Endpoint(
            settings={
                'response_type': ([BillingCostBudgetRecord],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/running-cost/{field}',
                'operation_id': 'get_running_cost',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'field',
                    'invoice_month',
                    'source',
                ],
                'required': [
                    'field',
                ],
                'nullable': [
                    'invoice_month',
                    'source',
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'field':
                        (BillingColumn,),
                    'invoice_month':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                    'source':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                },
                'attribute_map': {
                    'field': 'field',
                    'invoice_month': 'invoice_month',
                    'source': 'source',
                },
                'location_map': {
                    'field': 'path',
                    'invoice_month': 'query',
                    'source': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_sequencing_groups_async = async_wrap(self.get_sequencing_groups)
        self.get_sequencing_groups_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/sequencing-groups',
                'operation_id': 'get_sequencing_groups',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_sequencing_types_async = async_wrap(self.get_sequencing_types)
        self.get_sequencing_types_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/sequencing-types',
                'operation_id': 'get_sequencing_types',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_skus_async = async_wrap(self.get_skus)
        self.get_skus_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/skus',
                'operation_id': 'get_skus',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                    'limit',
                    'offset',
                ],
                'required': [],
                'nullable': [
                    'offset',
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'limit':
                        (int,),
                    'offset':
                        (bool, date, datetime, dict, float, int, list, str, none_type,),
                },
                'attribute_map': {
                    'limit': 'limit',
                    'offset': 'offset',
                },
                'location_map': {
                    'limit': 'query',
                    'offset': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_stages_async = async_wrap(self.get_stages)
        self.get_stages_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/stages',
                'operation_id': 'get_stages',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_topics_async = async_wrap(self.get_topics)
        self.get_topics_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/topics',
                'operation_id': 'get_topics',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_total_cost_async = async_wrap(self.get_total_cost)
        self.get_total_cost_endpoint = _Endpoint(
            settings={
                'response_type': ([BillingTotalCostRecord],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/total-cost',
                'operation_id': 'get_total_cost',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'billing_total_cost_query_model',
                ],
                'required': [
                    'billing_total_cost_query_model',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'billing_total_cost_query_model':
                        (BillingTotalCostQueryModel,),
                },
                'attribute_map': {
                },
                'location_map': {
                    'billing_total_cost_query_model': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.get_wdl_task_names_async = async_wrap(self.get_wdl_task_names)
        self.get_wdl_task_names_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/billing/wdl-task-names',
                'operation_id': 'get_wdl_task_names',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.is_billing_enabled_async = async_wrap(self.is_billing_enabled)
        self.is_billing_enabled_endpoint = _Endpoint(
            settings={
                'response_type': (bool,),
                'auth': [],
                'endpoint_path': '/api/v1/billing/is-billing-enabled',
                'operation_id': 'is_billing_enabled',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

    def cost_by_ar_guid(
        self,
        ar_guid,
        **kwargs
    ):
        """Get Cost By Ar Guid  # noqa: E501

        Get Hail Batch costs by AR GUID  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.cost_by_ar_guid(ar_guid, async_req=True)
        >>> result = thread.get()

        Args:
            ar_guid (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [AnalysisCostRecord]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['ar_guid'] = \
            ar_guid
        return self.cost_by_ar_guid_endpoint.call_with_http_info(**kwargs)

    def cost_by_batch_id(
        self,
        batch_id,
        **kwargs
    ):
        """Get Cost By Batch Id  # noqa: E501

        Get Hail Batch costs by Batch ID  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.cost_by_batch_id(batch_id, async_req=True)
        >>> result = thread.get()

        Args:
            batch_id (str):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [AnalysisCostRecord]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['batch_id'] = \
            batch_id
        return self.cost_by_batch_id_endpoint.call_with_http_info(**kwargs)

    def get_compute_categories(
        self,
        **kwargs
    ):
        """Get Compute Categories  # noqa: E501

        Get list of all compute categories in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_compute_categories(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_compute_categories_endpoint.call_with_http_info(**kwargs)

    def get_cost_categories(
        self,
        **kwargs
    ):
        """Get Cost Categories  # noqa: E501

        Get list of all service description / cost categories in database  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_cost_categories(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_cost_categories_endpoint.call_with_http_info(**kwargs)

    def get_cromwell_sub_workflow_names(
        self,
        **kwargs
    ):
        """Get Cromwell Sub Workflow Names  # noqa: E501

        Get list of all cromwell_sub_workflow_names in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_cromwell_sub_workflow_names(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_cromwell_sub_workflow_names_endpoint.call_with_http_info(**kwargs)

    def get_datasets(
        self,
        **kwargs
    ):
        """Get Datasets  # noqa: E501

        Get list of all datasets in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_datasets(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_datasets_endpoint.call_with_http_info(**kwargs)

    def get_gcp_projects(
        self,
        **kwargs
    ):
        """Get Gcp Projects  # noqa: E501

        Get list of all GCP projects in database  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_gcp_projects(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_gcp_projects_endpoint.call_with_http_info(**kwargs)

    def get_invoice_months(
        self,
        **kwargs
    ):
        """Get Invoice Months  # noqa: E501

        Get list of all invoice months in database Results are sorted DESC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_invoice_months(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_invoice_months_endpoint.call_with_http_info(**kwargs)

    def get_namespaces(
        self,
        **kwargs
    ):
        """Get Namespaces  # noqa: E501

        Get list of all namespaces in database Results are sorted DESC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_namespaces(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_namespaces_endpoint.call_with_http_info(**kwargs)

    def get_running_cost(
        self,
        field,
        **kwargs
    ):
        """Get Running Costs  # noqa: E501

        Get running cost for specified fields in database e.g. fields = ['gcp_project', 'topic', 'wdl_task_names', 'cromwell_sub_workflow_name', 'compute_category']  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_running_cost(field, async_req=True)
        >>> result = thread.get()

        Args:
            field (BillingColumn):

        Keyword Args:
            invoice_month (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            source (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [BillingCostBudgetRecord]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['field'] = \
            field
        return self.get_running_cost_endpoint.call_with_http_info(**kwargs)

    def get_sequencing_groups(
        self,
        **kwargs
    ):
        """Get Sequencing Groups  # noqa: E501

        Get list of all sequencing_groups in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_sequencing_groups(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_sequencing_groups_endpoint.call_with_http_info(**kwargs)

    def get_sequencing_types(
        self,
        **kwargs
    ):
        """Get Sequencing Types  # noqa: E501

        Get list of all sequencing_types in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_sequencing_types(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_sequencing_types_endpoint.call_with_http_info(**kwargs)

    def get_skus(
        self,
        **kwargs
    ):
        """Get Skus  # noqa: E501

        Get list of all SKUs in database There is over 400 Skus so limit is required Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_skus(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            limit (int): [optional] if omitted the server will use the default value of 10
            offset (bool, date, datetime, dict, float, int, list, str, none_type): [optional]
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_skus_endpoint.call_with_http_info(**kwargs)

    def get_stages(
        self,
        **kwargs
    ):
        """Get Stages  # noqa: E501

        Get list of all stages in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_stages(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_stages_endpoint.call_with_http_info(**kwargs)

    def get_topics(
        self,
        **kwargs
    ):
        """Get Topics  # noqa: E501

        Get list of all topics in database  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_topics(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_topics_endpoint.call_with_http_info(**kwargs)

    def get_total_cost(
        self,
        billing_total_cost_query_model,
        **kwargs
    ):
        """Get Total Cost  # noqa: E501

        Get Total cost of selected fields for requested time interval  Here are few examples of requests:  1. Get total topic for month of March 2023, ordered by cost DESC:      {         \"fields\": [\"topic\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true},         \"source\": \"aggregate\"     }  2. Get total cost by day and topic for March 2023, order by day ASC and topic DESC:      {         \"fields\": [\"day\", \"topic\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-05\",         \"order_by\": {\"day\": false, \"topic\": true},         \"limit\": 10     }  3. Get total cost of sku and cost_category for topic 'hail', March 2023,    order by cost DESC with Limit and Offset:      {         \"fields\": [\"sku\", \"cost_category\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"filters\": { \"topic\": \"hail\"},         \"order_by\": {\"cost\": true},         \"limit\": 5,         \"offset\": 10     }  4. Get total cost per dataset for month of March, order by cost DESC:      {         \"fields\": [\"dataset\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true}     }  5. Get total cost of daily cost for dataset 'acute-care', March 2023,    order by day ASC and dataset DESC:      {         \"fields\": [\"day\", \"dataset\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"filters\": { \"dataset\": \"acute-care\"},         \"order_by\": {\"day\": false, \"dataset\": true}     }  6. Get total cost for given batch_id and category for month of March,    order by cost DESC:      {         \"fields\": [\"cost_category\", \"batch_id\", \"sku\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"filters\": { \"batch_id\": \"423094\"},         \"order_by\": {\"cost\": true},         \"limit\": 5     }  7. Get total sequencing_type for month of March 2023, ordered by cost DESC:      {         \"fields\": [\"sequencing_type\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true}     }  8. Get total stage for month of March 2023, ordered by cost DESC:      {         \"fields\": [\"stage\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true}     }  9. Get total sequencing_group for month of March 2023, ordered by cost DESC:      {         \"fields\": [\"sequencing_group\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true}     }  10. Get total gcp_project for month of March 2023, ordered by cost DESC:      {         \"fields\": [\"gcp_project\"],         \"start_date\": \"2023-03-01\",         \"end_date\": \"2023-03-31\",         \"order_by\": {\"cost\": true},         \"source\": \"gcp_billing\"     }  11. Get total cost by sku for given ar_guid, order by cost DESC:      {         \"fields\": [\"sku\"],         \"start_date\": \"2023-10-23\",         \"end_date\": \"2023-10-23\",         \"filters\": { \"ar_guid\": \"4e53702e-8b6c-48ea-857f-c5d33b7e72d7\"},         \"order_by\": {\"cost\": true}     }  12. Get total cost by compute_category order by cost DESC:      {         \"fields\": [\"compute_category\"],         \"start_date\": \"2023-11-10\",         \"end_date\": \"2023-11-10\",         \"order_by\": {\"cost\": true}     }  13. Get total cost by cromwell_sub_workflow_name, order by cost DESC:      {         \"fields\": [\"cromwell_sub_workflow_name\"],         \"start_date\": \"2023-11-10\",         \"end_date\": \"2023-11-10\",         \"order_by\": {\"cost\": true}     }  14. Get total cost by sku for given cromwell_workflow_id, order by cost DESC:      {         \"fields\": [\"sku\"],         \"start_date\": \"2023-11-10\",         \"end_date\": \"2023-11-10\",         \"filters\": {\"cromwell_workflow_id\": \"cromwell-00448f7b-8ef3-4d22-80ab-e302acdb2d28\"},         \"order_by\": {\"cost\": true}     }  15. Get total cost by sku for given goog_pipelines_worker, order by cost DESC:      {         \"fields\": [\"goog_pipelines_worker\"],         \"start_date\": \"2023-11-10\",         \"end_date\": \"2023-11-10\",         \"order_by\": {\"cost\": true}     }  16. Get total cost by sku for given wdl_task_name, order by cost DESC:      {         \"fields\": [\"wdl_task_name\"],         \"start_date\": \"2023-11-10\",         \"end_date\": \"2023-11-10\",         \"order_by\": {\"cost\": true}     }  17. Get total cost by sku for provided ID, which can be any of [ar_guid, batch_id, sequencing_group or cromwell_workflow_id], order by cost DESC:      {         \"fields\": [\"sku\", \"ar_guid\", \"batch_id\", \"sequencing_group\", \"cromwell_workflow_id\"],         \"start_date\": \"2023-11-01\",         \"end_date\": \"2023-11-30\",         \"filters\": {             \"ar_guid\": \"855a6153-033c-4398-8000-46ed74c02fe8\",             \"batch_id\": \"429518\",             \"sequencing_group\": \"cpg246751\",             \"cromwell_workflow_id\": \"cromwell-e252f430-4143-47ec-a9c0-5f7face1b296\"         },         \"filters_op\": \"OR\",         \"order_by\": {\"cost\": true}     }  18. Get weekly total cost by sku for selected cost_category, order by day ASC:      {         \"fields\": [\"sku\"],         \"start_date\": \"2022-11-01\",         \"end_date\": \"2023-12-07\",         \"filters\": {             \"cost_category\": \"Cloud Storage\"         },         \"order_by\": {\"day\": false},         \"time_periods\": \"week\"     }  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_total_cost(billing_total_cost_query_model, async_req=True)
        >>> result = thread.get()

        Args:
            billing_total_cost_query_model (BillingTotalCostQueryModel):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [BillingTotalCostRecord]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['billing_total_cost_query_model'] = \
            billing_total_cost_query_model
        return self.get_total_cost_endpoint.call_with_http_info(**kwargs)

    def get_wdl_task_names(
        self,
        **kwargs
    ):
        """Get Wdl Task Names  # noqa: E501

        Get list of all wdl_task_names in database Results are sorted ASC  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_wdl_task_names(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_wdl_task_names_endpoint.call_with_http_info(**kwargs)

    def is_billing_enabled(
        self,
        **kwargs
    ):
        """Is Billing Enabled  # noqa: E501

        Return true if billing ie enabled, false otherwise  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.is_billing_enabled(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.is_billing_enabled_endpoint.call_with_http_info(**kwargs)




import dataclasses
import unittest

from db.python.filters import GenericFilter, GenericFilterModel


@dataclasses.dataclass(kw_only=True)
class GenericFilterTest(GenericFilterModel):
    """Test model for GenericFilter"""

    test_string: GenericFilter[str] | None = None
    test_int: GenericFilter[int] | None = None
    test_dict: dict[str, GenericFilter[str]] | None = None


class TestGenericFilters(unittest.TestCase):
    """Test generic filters SQL generation"""

    def test_basic_no_override(self):
        """Test that the basic filter converts to SQL as expected"""
        filter_ = GenericFilterTest(test_string=GenericFilter(eq='test'))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string = :test_string_eq', sql)
        self.assertDictEqual({'test_string_eq': 'test'}, values)

    def test_contains_case_sensitive(self):
        """Test that the basic filter converts to SQL as expected"""
        filter_ = GenericFilterTest(test_string=GenericFilter(contains='test'))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string LIKE :test_string_contains', sql)
        self.assertDictEqual({'test_string_contains': '%test%'}, values)

    def test_malicious_contains(self):
        """Test that a contains-% filter converts to SQL by appropriately encoding _ and %"""
        filter_ = GenericFilterTest(test_string=GenericFilter(contains='per%ce_nt'))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string LIKE :test_string_contains', sql)
        self.assertDictEqual({'test_string_contains': r'%per\%ce\_nt%'}, values)

    def test_icontains_is_not_case_sensitive(self):
        """Test that the basic filter converts to SQL as expected"""
        filter_ = GenericFilterTest(test_string=GenericFilter(icontains='test'))
        sql, values = filter_.to_sql()

        self.assertEqual('LOWER(test_string) LIKE LOWER(:test_string_icontains)', sql)
        self.assertDictEqual({'test_string_icontains': '%test%'}, values)

    def test_malicious_icontains(self):
        """Test that an icontains-% filter converts to SQL by appropriately encoding _ and %"""
        filter_ = GenericFilterTest(test_string=GenericFilter(icontains='per%ce_nt'))
        sql, values = filter_.to_sql()

        self.assertEqual('LOWER(test_string) LIKE LOWER(:test_string_icontains)', sql)
        self.assertDictEqual({'test_string_icontains': r'%per\%ce\_nt%'}, values)

    def test_basic_override(self):
        """Test that the basic filter with an override converts to SQL as expected"""
        filter_ = GenericFilterTest(test_string=GenericFilter(eq='test'))
        sql, values = filter_.to_sql({'test_string': 't.string'})

        self.assertEqual('t.string = :t_string_eq', sql)
        self.assertDictEqual({'t_string_eq': 'test'}, values)

    def test_in_single(self):
        """
        Test that a single value filtered using the "in" operator
        gets converted to an eq operation
        """
        filter_ = GenericFilterTest(test_string=GenericFilter(in_=['test']))
        sql, values = filter_.to_sql()

        self.assertEqual('test_string = :test_string_in_eq', sql)
        self.assertDictEqual({'test_string_in_eq': 'test'}, values)

    def test_in_multiple(self):
        """
        Test that values filtered using the "in" operator convert as expected
        """
        value = [1, 2]
        filter_ = GenericFilterTest(test_int=GenericFilter(in_=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int IN :test_int_in', sql)
        self.assertDictEqual({'test_int_in': value}, values)

    def test_gt_single(self):
        """
        Test that a single value filtered using the "gt" operator
        """
        filter_ = GenericFilterTest(test_int=GenericFilter(gt=123))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int > :test_int_gt', sql)
        self.assertDictEqual({'test_int_gt': 123}, values)

    def test_gte_single(self):
        """
        Test that a single value filtered using the "gte" operator
        """
        filter_ = GenericFilterTest(test_int=GenericFilter(gte=123))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int >= :test_int_gte', sql)
        self.assertDictEqual({'test_int_gte': 123}, values)

    def test_lt_single(self):
        """
        Test that a single value filtered using the "lt" operator
        """
        filter_ = GenericFilterTest(test_int=GenericFilter(lt=123))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int < :test_int_lt', sql)
        self.assertDictEqual({'test_int_lt': 123}, values)

    def test_lte_single(self):
        """
        Test that a single value filtered using the "lte" operator
        """
        filter_ = GenericFilterTest(test_int=GenericFilter(lte=123))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int <= :test_int_lte', sql)
        self.assertDictEqual({'test_int_lte': 123}, values)

    def test_not_in_multiple(self):
        """
        Test that values filtered using the "nin" operator convert as expected
        """
        value = [1, 2]
        filter_ = GenericFilterTest(test_int=GenericFilter(nin=value))
        sql, values = filter_.to_sql()

        self.assertEqual('test_int NOT IN :test_int_nin', sql)
        self.assertDictEqual({'test_int_nin': value}, values)
